#! /usr/bin/python3

try :
    import sys, os
    from os.path import basename, dirname
    import parser, symbol, token, types
    import subprocess
except :
    raise
else :
    def _(s) : return s

    rpm_br=os.path.normpath(os.getenv('RPM_BUILD_ROOT',""))
    if rpm_br in ['','/','.'] :
        rpm_br=''
        br_len = 0
    else :
        br_len = len(rpm_br)
    #print rpm_br
    def strip_buildroot(p) :
        if p.startswith(rpm_br+'/') :
                return p[br_len:]
        return p

    prefix = "python%u" % sys.version_info[0]
    def requires(m) :
        req = "%s(%s)" % (prefix,m)
        print(req)

    symbol_skip = [symbol.encoding_decl, symbol.classdef]

    def has1(x) :
        return (type(x) in [list, tuple]) and (len(x) > 1)
    def pro(x) :
        return x[1]
    def namelist(import_node) :
        return [x for x in import_node[2][1:] if x[0] != 12 ]

    # Provides: from python3-base package
    base_modules = []
    # disabling for bootstrap
    # base_provides = subprocess.check_output(['rpmquery','--provides', 'python3-base']).split(b'\n')
    # for module in base_provides:
    #    if module.strip()[:9] == prefix :
    #        base_modules.append(module.strip()[10:][:-1])

    def ignore(module) :
        if module in dict([ (x,1) for x in list(os.getenv('RPM_PYTHON3_REQ_SKIP',"").split()) + list(sys.builtin_module_names)]) :
            return True
        else :
            return False

    REQ=os.getenv('RPM_PYTHON3_REQ_METHOD','strict')
    IS_HIER = os.getenv('RPM_PYTHON3_REQ_HIER',None)
    if IS_HIER and IS_HIER.startswith('no') :
        # rpm-build-python/tools/requires.py uses 'hier' as the default value;
        # rpm-build-python3/python3 can set it to 'yes'.
        # So, let's suppose the opposite value can be either 'nohier' or 'no'
        # (or simply an unset env variable). Here we map those possibilities
        # to our internal canonical representation of this value
        # (just in case someone may use them).
        IS_HIER = None

    src = None
    def match(tree,deep=0) :
        if tree[0] not in symbol_skip :
            deep += 1
        for node in tree :
            if has1(node) :
                line = 0
                deps = []
                if node[0] == symbol.import_stmt :
                    node = pro(node)
                    if node[1][1] == 'import' :
                        line = node[1][2]
                        for name in namelist(node) :
                            if IS_HIER is None :
                                deps.append(name[1][1][1])
                            else :
                                deps.append(".".join( [ i[1] for i in  name[1][1:] if i[0]==1 ]))
                    elif node[1][1] == 'from' :
                        line = node[1][2]
                        if type(node[2][1]) is list:
                            if IS_HIER is None :
                                deps.append(node[2][1][1])
                            else :
                                deps.append(".".join( [ i[1] for i in  node[2][1:] if i[0]==1 ]))
                        else:   # XXX temporary 2.5 PEP 328 workaround
                            print("%s: %s: line=%d possible relative import from %s, UNIMPLEMENTED" % (sys.argv[0], src, line, node[2][1]), file=sys.stderr)
                else :
                    #print(node, file=sys.stderr) # DEBUG
                    # I've got some minimal sufficient understanding of the Python grammar
                    # from the excerpt at http://effbot.org/zone/simple-top-down-parser.htm
                    # --imz
                    def maybe_unpack(node, symbol_type) :
                        if has1(node) and (node[0] == symbol_type) :
                            return node[1]
                        else :
                            return None
                    def must_unpack(node, symbol_type) :
                        if has1(node) :
                            assert (node[0] == symbol_type)
                            return node[1]
                        else :
                            return None
                    def maybe_unpack_until(node, symbol_type) :
                        curr_node = node
                        while has1(curr_node) and (curr_node[0] != symbol.atom) :
                            curr_node = pro(curr_node)
                        return maybe_unpack(curr_node, symbol_type)
                    def maybe_unpack_atomWithTrailerNode(node) :
                        if has1(node) :
                            atom = maybe_unpack(node[1], symbol.atom)
                            if atom and (len(node) > 2) and has1(node[2]) :
                                # Since Python 3.5, atom is an immediate constituent of atom_expr,
                                # which can only have a trailer as the next sister node.
                                # Python 3.5's grammar is nice.
                                # We'd like to code according to it.
                                #
                                # The condition below is an exception for Python 3.3 grammar,
                                # which had no atom_expr, and there the next sister node
                                # could have a mix of types.
                                if (node[0] == symbol.power) and (node[2][0] != symbol.trailer) :
                                    # Python < 3.5
                                    return (atom, None)
                                assert (node[2][0] == symbol.trailer)
                                return (atom, node[2])
                            else :
                                return (atom, None)
                        else :
                            return (None, None)
                    def maybe_unpack_until_atomWithTrailerNode(node) :
                        curr_node = node
                        while has1(curr_node) :
                            (atom, trailer_node) = maybe_unpack_atomWithTrailerNode(curr_node)
                            if atom :
                                return (atom, trailer_node)
                            curr_node = pro(curr_node)
                        return (None,None)

                    (atom, trailer_node) = maybe_unpack_until_atomWithTrailerNode(node)
                    if atom and (atom[1] == '__import__'):
                        line = atom[2]
                        if trailer_node :
                            if (trailer_node[1][1] == '(') :
                                arglist1 = maybe_unpack(trailer_node[2], symbol.arglist)
                                arg_atom = maybe_unpack_until(arglist1, symbol.atom)
                                if has1(arg_atom) :
                                    def unquote(s) :
                                        if (len(s) > 1) and (s[0] in ["'",'"']) and (s[-1] == s[0]) :
                                            return s[1:-1]
                                        else :
                                            return None
                                    m = unquote(arg_atom[1])
                                    if m :
                                        deps.append(m)
                                    else :
                                        print('%s: %s: line=%d UNABLE to determine the imported name at parse-time: %s' % (sys.argv[0], src, line, arg_atom[1]), file=sys.stderr)
                        if [] == deps :
                            print('%s: %s: line=%d a form of __import__ was read which gave no deps: %s' % (sys.argv[0], src, line, node), file=sys.stderr)

                if REQ not in ['slight','relaxed'] or deep <= 4 :
                    for dep in deps :
                        yield dep
                else :
                    for dep in deps:
                        print("%s: %s: line=%d IGNORE (for REQ=%s and deep=%d) module=%s" % (sys.argv[0], src, line, REQ, deep, dep), file=sys.stderr)

                for item in match(node,deep) :
                    yield item

    if __name__ == '__main__' :
        files = sys.argv[1:] or [ x.strip() for x in sys.stdin.readlines() ]

        reqs = {}
        prov = {}
        for src in files :
            prov.setdefault(dirname(src),[]).append(os.path.splitext(basename(src))[0])
            if basename(src) == "__init__.py" :
                prov.setdefault(dirname(dirname(src)),[]).append(basename(dirname(src)))


        prov_new = {}
        for d,reqs in list(prov.items()) :
            l = []
            r = d.split("/")
            while r :
                item = r.pop()
                l.insert(0,item)
                for req in reqs :
                    prov_new.setdefault("/".join(r),[]).append(".".join(l + [req]))

        for path,reqs in list(prov_new.items()) :
            prov.setdefault(path,[]).extend(reqs)

        #for d,reqs in prov.items() :
        #    print d
        #    for req in reqs :
        #        print "\t",req

        all_prov_items = []
        list(map(lambda items: all_prov_items.extend(items), list(prov.values())))

        import re
        # Parser/tokenizer.c:get_coding_spec()
        re_coding = re.compile(r'^\s*#.*?coding[:=]\s*([\w._-]+)')

        def get_raw_encoding(line) :
            enc = re_coding.search(line)
            if enc :
                return enc.group(1)

        # Parser/tokenizer.c:get_normal_name()
        def get_enc_normal_name(enc) :
            enc = enc.lower().replace('_','-')
            if enc.startswith('utf-8') :
                return 'utf-8'
            for alias in ['latin-1','iso-8859-1','iso-latin-1'] :
                if enc.startswith(alias) :
                    return 'iso-8859-1'
            return enc

        missing_encodings = None
        def get_encoding_module(enc) :
            mod = 'encodings'
            try :
                from encodings import search_function
            except :
                global missing_encodings
                missing_encodings = True
                return mod
            try :
                mod = search_function(enc)[1].__self__.__class__.__module__
                print("python3.req: %s: encoding=%s module=%s" % (src,enc,mod), file=sys.stderr)
            except :
                pass
            return mod

        def need_encoding_module(line):
            enc = get_raw_encoding(line)
            if enc :
                enc = get_enc_normal_name(enc)
                if enc not in ['utf-8','iso-8859-1'] :
                    print("python3.req: %s: non-standard encoding: %s" % (src,enc), file=sys.stderr)
                    mod = get_encoding_module(enc)
                    return mod

        found_deps = []

        for src in files :
            ext = os.path.splitext(basename(src))[1]
            if ext == '.so' :
                dep_version = os.getenv('RPM_PYTHON3_VERSION',
                                        "%s.%s" % sys.version_info[0:2])
                # Borrowed the method from lib.req:
                import subprocess
                dump = subprocess.Popen(["objdump", "-p", src],
                                        stdout=subprocess.PIPE)
                # TODO: rewrite the grep call in Python.
                if 0 == subprocess.call(["fgrep", "-q",
                                         "file format elf64"],
                                        stdin=dump.stdout) :
                    dep_suffix = "(64bit)"
                else :
                    dep_suffix = ""
                print("python%s-ABI%s" % (dep_version,dep_suffix))
            if ext not in ['.so','.pyc','.pyo','.pth'] :
                try:
                    lines = [ line.rstrip().replace('\r','') for line in open(src, encoding="utf-8").readlines() ]
                except:
                    lines = []
                encmod = None
                if len(lines) > 0 :
                    encmod = need_encoding_module(lines[0])
                if not encmod and len(lines) > 1 :
                    encmod = need_encoding_module(lines[1])
                if encmod :
                    requires(encmod)

                def must_fail() :
                    if ext == '.py' :
                        return True
                    if len(lines) > 0 and lines[0].startswith('#!') :
                        return True
                try :
                    lis = parser.suite('\n'.join(lines)+'\n').tolist(line_info=1)
                except Exception as msg :
                    print('python3.req: ERROR: %s: %s' % (src,msg), file=sys.stderr)
                    if encmod and missing_encodings :
                        print("python3.req: maybe you need python3-modules-encodings", file=sys.stderr)
                    if must_fail() :
                        raise
                    else :
                        print("python3.req: %s: possibly not pythonish file" % src, file=sys.stderr)
                        continue

                for item in match(lis) :
                    #if not item in prov[dirname(src)] :
                    if not item in all_prov_items :
                        if not ignore(item) :
                    	    # generating all (including python3-base) dependencies, maybe check against base_modules in future
                            found_deps.append(item)
                        else :
                            print("python3.req: %s: skipping %s" % (src, item), file=sys.stderr)
                    # else :
                    #     print("python3.req: %s: hiding %s, which SEEMS TO BE SELF-SATISFIED" % (src, item), file=sys.stderr)

        for dep in found_deps :
            requires(dep)

        modules = [ os.path.normpath(x) for x in os.getenv('RPM_PYTHON3_MODULE_DECLARED','').split() ]
        #print modules

        # it's better to use set here
        # it's better to remove pythons with version < 2.4 HA XEP to use set here and arround
        dirs = {}
        for src in files :
            #print 1,src
            if os.path.splitext(basename(src))[1] in ['.py',''] :
                src = dirname(os.path.normpath(src))
                #print 2,br_len,strip_buildroot(src)
                if strip_buildroot(src) in modules:
                    dirs[src] = 1

        #print dirs	
        # collected [unique] directories which are marked as separate packages.
        # now filter them to keep only those which don't have __init__.py (in this package)
        # and finally print required dependencies
        def print_func(a): print(a)
        
        list(map(lambda finit: print_func(strip_buildroot(finit)),                # print path to __init__.py w/o buildroot
            [init for init in [os.path.join(dir, '__init__.py') for dir in list(dirs.keys())] if not init in files]))

# ex: set ts=8 sts=4 sw=4 et:
